# --------------------------------------------------------
# Copyright (c) 2021 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Based on BEIT code bases (https://github.com/microsoft/unilm/tree/master/beit)
# Written by Yutong Lin, Zhenda Xie
# Modified by Yixuan Wei
# --------------------------------------------------------

import math
from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
import torch.utils.checkpoint as checkpoint
from timm.models.layers import DropPath, to_2tuple, trunc_normal_

class LinearFP32(nn.Linear):
    def __init__(self, in_features, out_features, bias=True):
        super(LinearFP32, self).__init__(in_features, out_features, bias)

    def forward(self, input: Tensor) -> Tensor:
        return F.linear(input.float(), self.weight.float(),
                        self.bias.float() if self.bias is not None else None)

class Mlp(nn.Module):
    def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
        super().__init__()
        out_features = out_features or in_features
        hidden_features = hidden_features or in_features
        self.fc1 = nn.Linear(in_features, hidden_features)
        self.act = act_layer()
        self.fc2 = nn.Linear(hidden_features, out_features)
        self.drop = nn.Dropout(drop)

    def forward(self, x):
        x = self.fc1(x)
        x = self.act(x)
        # x = self.drop(x)
        # comment out this for the orignal BERT implement
        x = self.fc2(x)
        x = self.drop(x)
        return x


class Attention(nn.Module):
    def __init__(
            self, dim, num_heads=8, qkv_bias=False, qk_scale=None, attn_drop=0.,
            proj_drop=0., window_size=None, attn_head_dim=None, with_cls_token=True,
            with_k_bias=False):
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        if attn_head_dim is not None:
            head_dim = attn_head_dim
        all_head_dim = head_dim * self.num_heads
        self.scale = qk_scale or head_dim ** -0.5

        self.qkv = nn.Linear(dim, all_head_dim * 3, bias=False)
        if qkv_bias:
            self.q_bias = nn.Parameter(torch.zeros(all_head_dim))
            self.v_bias = nn.Parameter(torch.zeros(all_head_dim))
        else:
            self.q_bias = None
            self.v_bias = None
        self.with_k_bias = with_k_bias
        if self.with_k_bias:
            self.k_bias = nn.Parameter(torch.zeros(all_head_dim))
        self.window_size = window_size
        if window_size:
            if with_cls_token:
                # extra 3: cls to token & token 2 cls & cls to cls
                self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1) + 3
                self.num_tokens = self.window_size[0] * self.window_size[1] + 1
            else:
                self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1)
                self.num_tokens = self.window_size[0] * self.window_size[1]
            self.relative_position_bias_table = nn.Parameter(
                torch.zeros(self.num_relative_distance, num_heads))  # 2*Wh-1 * 2*Ww-1, nH

            # get pair-wise relative position index for each token inside the window
            coords_h = torch.arange(window_size[0])
            coords_w = torch.arange(window_size[1])
            coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
            coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
            relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
            relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
            relative_coords[:, :, 0] += window_size[0] - 1  # shift to start from 0
            relative_coords[:, :, 1] += window_size[1] - 1
            relative_coords[:, :, 0] *= 2 * window_size[1] - 1

            if with_cls_token:
                relative_position_index = \
                    torch.zeros(size=(window_size[0] * window_size[1] + 1,) * 2, dtype=relative_coords.dtype)
                relative_position_index[1:, 1:] = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
                relative_position_index[0, 0:] = self.num_relative_distance - 3
                relative_position_index[0:, 0] = self.num_relative_distance - 2
                relative_position_index[0, 0] = self.num_relative_distance - 1
            else:
                relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww

            self.register_buffer("relative_position_index", relative_position_index)
        else:
            self.window_size = None
            self.relative_position_bias_table = None
            self.relative_position_index = None

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(all_head_dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x, rel_pos_bias=None, rpb_mask=None, attn_mask=None):
        B, N, C = x.shape
        qkv_bias = None
        if self.q_bias is not None:
            if self.with_k_bias:
                qkv_bias = torch.cat((self.q_bias, self.k_bias, self.v_bias))
            else:
                qkv_bias = torch.cat((self.q_bias, torch.zeros_like(self.v_bias, requires_grad=False), self.v_bias))
        qkv = F.linear(input=x, weight=self.qkv.weight, bias=qkv_bias)

        head_ft = qkv.reshape(B, N, 3, self.num_heads, -1).permute(0, 3, 1, 2, 4)

        qkv = qkv.reshape(B, N, 3, self.num_heads, -1).permute(2, 0, 3, 1, 4)
        q, k, v = qkv[0], qkv[1], qkv[2]   # make torchscript happy (cannot use tensor as tuple)

        
        q = q * self.scale
        attn = (q @ k.transpose(-2, -1))

        if self.relative_position_bias_table is not None:
            relative_position_bias = \
                self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
                    self.num_tokens, self.num_tokens, -1)  # Wh*Ww,Wh*Ww,nH
            relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous().unsqueeze(0)  # 1, nH, Wh*Ww, Wh*Ww

            if rpb_mask is not None:
                relative_position_bias = relative_position_bias.expand(B, -1, -1, -1)   # [B, nH, L + 1, L + 1]
                relative_position_bias = relative_position_bias[rpb_mask].reshape((B, -1, N, N))

            attn = attn + relative_position_bias

        if rel_pos_bias is not None:
            attn = attn + rel_pos_bias

        if attn_mask is not None:
            attn = attn - 1e10 * attn_mask

        attn = attn.softmax(dim=-1)
        attn = self.attn_drop(attn)

        x = (attn @ v).transpose(1, 2)
        x = x.reshape(B, N, -1)
        x = self.proj(x)
        x = self.proj_drop(x)
        # return x

        return x, attn, attn @ v, head_ft


class Block(nn.Module):

    def __init__(self, dim, num_heads, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop=0., attn_drop=0.,
                 drop_path=0., init_values=None, act_layer=nn.GELU, norm_layer=nn.LayerNorm,
                 window_size=None, attn_head_dim=None, with_cls_token=True,
                 with_k_bias=False):
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, qk_scale=qk_scale,
            attn_drop=attn_drop, proj_drop=drop, window_size=window_size, attn_head_dim=attn_head_dim,
            with_cls_token=with_cls_token, with_k_bias=with_k_bias)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
        self.norm2 = norm_layer(dim)
        mlp_hidden_dim = int(dim * mlp_ratio)
        self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

        if init_values is not None:
            self.gamma_1 = nn.Parameter(init_values * torch.ones((dim)),requires_grad=True)
            self.gamma_2 = nn.Parameter(init_values * torch.ones((dim)),requires_grad=True)
        else:
            self.gamma_1, self.gamma_2 = None, None

    def forward(self, x, rel_pos_bias=None, rpb_mask=None, attn_mask=None):
        attn_x, attn, attn_ft, head_ft = self.attn(self.norm1(x), rel_pos_bias=rel_pos_bias, rpb_mask=rpb_mask, attn_mask=attn_mask)
        if self.gamma_1 is None:
            x = x + self.drop_path(attn_x)
            x1 = x
            x = x + self.drop_path(self.mlp(self.norm2(x)))
            x2 = x
        else:
            x = x + self.drop_path(self.gamma_1 * attn_x)
            x1 = x
            x = x + self.drop_path(self.gamma_2 * self.mlp(self.norm2(x)))
            x2 = x
        return x, x1, x2, attn, attn_ft, head_ft


class PatchEmbed(nn.Module):
    """ Image to Patch Embedding
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, embed_dim=768):
        super().__init__()
        img_size = to_2tuple(img_size)
        patch_size = to_2tuple(patch_size)
        num_patches = (img_size[1] // patch_size[1]) * (img_size[0] // patch_size[0])
        self.patch_shape = (img_size[0] // patch_size[0], img_size[1] // patch_size[1])
        self.img_size = img_size
        self.patch_size = patch_size
        self.num_patches = num_patches

        self.proj = nn.Conv2d(in_chans, embed_dim, kernel_size=patch_size, stride=patch_size)

    def forward(self, x, **kwargs):
        B, C, H, W = x.shape
        # FIXME look at relaxing size constraints
        assert H == self.img_size[0] and W == self.img_size[1], \
            f"Input image size ({H}*{W}) doesn't match model ({self.img_size[0]}*{self.img_size[1]})."
        x = self.proj(x).flatten(2).transpose(1, 2)
        return x


class RelativePositionBias(nn.Module):

    def __init__(self, window_size, num_heads, with_cls_token=True):
        super().__init__()
        self.window_size = window_size
        if with_cls_token:
            # extra 3: cls to token & token 2 cls & cls to cls
            self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1) + 3
            self.num_tokens = self.window_size[0] * self.window_size[1] + 1
        else:
            self.num_relative_distance = (2 * window_size[0] - 1) * (2 * window_size[1] - 1)
            self.num_tokens = self.window_size[0] * self.window_size[1]
        self.relative_position_bias_table = nn.Parameter(
            torch.zeros(self.num_relative_distance, num_heads))  # 2*Wh-1 * 2*Ww-1, nH

        # get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(window_size[0])
        coords_w = torch.arange(window_size[1])
        coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += window_size[0] - 1  # shift to start from 0
        relative_coords[:, :, 1] += window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * window_size[1] - 1

        if with_cls_token:
            relative_position_index = \
                torch.zeros(size=(window_size[0] * window_size[1] + 1,) * 2, dtype=relative_coords.dtype)
            relative_position_index[1:, 1:] = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
            relative_position_index[0, 0:] = self.num_relative_distance - 3
            relative_position_index[0:, 0] = self.num_relative_distance - 2
            relative_position_index[0, 0] = self.num_relative_distance - 1
        else:
            relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww

        self.register_buffer("relative_position_index", relative_position_index)

    def mask_rel_pos_bias(self, rel_pos_bias, ids_keep):
        # rel_pos_bias: [num_heads, full_seq_len, full_seq_len]
        # ids_keep: [batch_size, num_tokens_to_keep]
        num_heads = self.relative_position_bias_table.shape[1]

        # 将ids_keep增加一个维度，以便将其应用于rel_pos_bias
        ids_keep_with_cls = torch.cat([torch.zeros(ids_keep.size(0), 1, dtype=torch.long, device=ids_keep.device), ids_keep + 1], dim=1)

        # 用gather选取rel_pos_bias中要保留的列
        rel_pos_bias_adjusted = torch.gather(rel_pos_bias.unsqueeze(0).expand(ids_keep.size(0), -1, -1, -1), dim=3, index=ids_keep_with_cls.unsqueeze(1).unsqueeze(2).expand(-1, num_heads, rel_pos_bias.size(-1), -1))

        rel_pos_bias_adjusted = torch.gather(rel_pos_bias_adjusted.permute(0,1,3,2), dim=3, index=ids_keep_with_cls.unsqueeze(1).unsqueeze(2).expand(-1, num_heads, ids_keep_with_cls.size(-1), -1))
        rel_pos_bias_adjusted = rel_pos_bias_adjusted.permute(0,1,3,2)

        return rel_pos_bias_adjusted


    def forward(self, ids_keep=None):
        relative_position_bias = \
            self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
                self.num_tokens, self.num_tokens, -1)  # Wh*Ww,Wh*Ww,nH
        relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww

        if ids_keep is not None:
            relative_position_bias_masked = self.mask_rel_pos_bias(relative_position_bias, ids_keep)
            relative_position_bias = relative_position_bias_masked

        return relative_position_bias


class VisionTransformer(nn.Module):
    """ Vision Transformer with support for patch or hybrid CNN input stage
    """
    def __init__(self, img_size=224, patch_size=16, in_chans=3, num_classes=1000, embed_dim=768, depth=12,
                 num_heads=12, mlp_ratio=4., qkv_bias=False, qk_scale=None, drop_rate=0., attn_drop_rate=0.,
                 drop_path_rate=0., norm_layer=nn.LayerNorm, init_values=None,
                 use_abs_pos_emb=True, use_rel_pos_bias=False, use_shared_rel_pos_bias=False, 
                 use_mean_pooling=True, init_scale=0.001, with_cls_token=True, use_checkpoint=False, with_k_bias=False):
        super().__init__()
        self.num_classes = num_classes
        self.num_features = self.embed_dim = embed_dim
        self.patch_size = patch_size
        self.in_chans = in_chans
        self.num_heads = num_heads
        self.use_abs_pos_emb = use_abs_pos_emb
        self.with_cls_token = with_cls_token
        self.use_checkpoint = use_checkpoint

        self.patch_embed = PatchEmbed(
            img_size=img_size, patch_size=patch_size, in_chans=in_chans, embed_dim=embed_dim)
        num_patches = self.patch_embed.num_patches
        self.num_patches = num_patches

        if with_cls_token:
            self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        else:
            self.cls_token = None

        if use_abs_pos_emb:
            self.pos_embed = nn.Parameter(torch.zeros(1, num_patches + 1, embed_dim)) if with_cls_token else nn.Parameter(torch.zeros(1, num_patches, embed_dim))
        else:
            self.pos_embed = None
        self.pos_drop = nn.Dropout(p=drop_rate)
        
        if use_shared_rel_pos_bias:
            self.rel_pos_bias = RelativePositionBias(window_size=self.patch_embed.patch_shape, num_heads=num_heads,
                                                     with_cls_token=with_cls_token)
        else:
            self.rel_pos_bias = None

        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, depth)]  # stochastic depth decay rule
        self.use_rel_pos_bias = use_rel_pos_bias
        self.blocks = nn.ModuleList([
            Block(
                dim=embed_dim, num_heads=num_heads, mlp_ratio=mlp_ratio, qkv_bias=qkv_bias, qk_scale=qk_scale,
                drop=drop_rate, attn_drop=attn_drop_rate, drop_path=dpr[i], norm_layer=norm_layer,
                init_values=init_values, window_size=self.patch_embed.patch_shape if use_rel_pos_bias else None,
                with_cls_token=with_cls_token, with_k_bias=with_k_bias)
            for i in range(depth)])
        self.norm = nn.Identity() if use_mean_pooling else norm_layer(embed_dim)
        self.fc_norm = norm_layer(embed_dim) if use_mean_pooling else None
        if num_classes > 0:
            self.head = nn.Linear(embed_dim, num_classes)
        else:
            self.head = nn.Identity()

        if self.pos_embed is not None:
            self._trunc_normal_(self.pos_embed, std=.02)
        if self.cls_token is not None:
            self._trunc_normal_(self.cls_token, std=.02)
        if num_classes > 0:
            self._trunc_normal_(self.head.weight, std=.02)
        self.apply(self._init_weights)
        self.fix_init_weight()

        if num_classes > 0:
            self.head.weight.data.mul_(init_scale)
            self.head.bias.data.mul_(init_scale)


    def _trunc_normal_(self, tensor, mean=0., std=1.):
        trunc_normal_(tensor, mean=mean, std=std)

    def fix_init_weight(self):
        def rescale(param, layer_id):
            param.div_(math.sqrt(2.0 * layer_id))

        for layer_id, layer in enumerate(self.blocks):
            rescale(layer.attn.proj.weight.data, layer_id + 1)
            rescale(layer.mlp.fc2.weight.data, layer_id + 1)

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            self._trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)
        elif isinstance(m, nn.Conv2d):
            self._trunc_normal_(m.weight, std=.02)
            if m.bias is not None:
                nn.init.constant_(m.bias, 0)

    def get_num_layers(self):
        return len(self.blocks)

    @torch.jit.ignore
    def no_weight_decay(self):
        return {'pos_embed', 'cls_token'}

    def get_classifier(self):
        return self.head

    def reset_classifier(self, num_classes, global_pool=''):
        self.num_classes = num_classes
        self.head = nn.Linear(self.embed_dim, num_classes) if num_classes > 0 else nn.Identity()

    def forward_features(self, x):
        zs = []
        attns = []
        attn_fts = []
        head_fts = []

        x = self.patch_embed(x)
        batch_size, seq_len, _ = x.size()

        if self.cls_token is not None:
            cls_tokens = self.cls_token.expand(batch_size, -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
            x = torch.cat((cls_tokens, x), dim=1)
        if self.pos_embed is not None:
            x = x + self.pos_embed
        x = self.pos_drop(x)

        zs.append(x)

        rel_pos_bias = self.rel_pos_bias() if self.rel_pos_bias is not None else None
        for blk in self.blocks:
            if self.use_checkpoint:
                x, x1, x2, attn, attn_ft, head_ft = checkpoint.checkpoint(blk, x, rel_pos_bias)
            else:
                x, x1, x2, attn, attn_ft, head_ft = blk(x, rel_pos_bias=rel_pos_bias)
            
            zs.append(x1)
            zs.append(x2)
            attns.append(attn)
            attn_fts.append(attn_ft)
            head_fts.append(head_ft)

        x = self.norm(x)
        if self.fc_norm is not None:
            t = x[:, 1:, :] if self.cls_token is not None else x
            return self.fc_norm(t.mean(1)), zs, attns, attn_fts,head_fts
        else:
            if self.cls_token is not None:
                return x[:, 0], zs, attns, attn_fts,head_fts
            else:
                raise ValueError

    def forward(self, x):
        x, zs, attns, attn_fts,head_fts = self.forward_features(x)
        x = self.head(x)

        zs.append(x)
        return x, zs, attns, attn_fts, head_fts

def vit_base(**kwargs):
    assert 'use_abs_pos_emb' in kwargs.keys(), 'use_abs_pos_emb must be specified'
    assert 'use_shared_rel_pos_bias' in kwargs.keys(), 'use_shared_rel_pos_bias must be specified'

    init_values = 0.1 # 没有修改的默认参数，来自FD的代码
    
    if 'init_values' in kwargs:
        import warnings
        warnings.warn("你已经修改了init_values, 这会导致transformer block没有gamma")

        init_values = kwargs['init_values']

    model = VisionTransformer(
        img_size=224, patch_size=16, num_classes=1000, embed_dim=768, depth=12, num_heads=12, mlp_ratio=4, qkv_bias=True,
        with_k_bias=False, drop_rate=0.0, drop_path_rate=0.0, init_values=init_values, use_abs_pos_emb=kwargs.get('use_abs_pos_emb', True),
        use_rel_pos_bias=kwargs.get('use_rel_pos_bias', False), use_shared_rel_pos_bias = kwargs.get('use_shared_rel_pos_bias', False),
        with_cls_token = True, use_checkpoint=kwargs.get('use_checkpoint', False), norm_layer=partial(nn.LayerNorm, eps=1e-6))

    
    return model

def build_vit(config):
    model = VisionTransformer(
        img_size=config.DATA.IMG_SIZE, # 224
        patch_size=config.MODEL.VIT.PATCH_SIZE,  # 16
        in_chans=config.MODEL.VIT.IN_CHANS,  # 3
        num_classes=config.MODEL.NUM_CLASSES,  # 0
        embed_dim=config.MODEL.VIT.EMBED_DIM,  # 768
        depth=config.MODEL.VIT.DEPTH,  # 12
        num_heads=config.MODEL.VIT.NUM_HEADS,  # 12
        mlp_ratio=config.MODEL.VIT.MLP_RATIO,  # 4.
        qkv_bias=config.MODEL.VIT.QKV_BIAS,  # False
        with_k_bias=config.DEV.VIT_WITHKBIAS,  # False
        drop_rate=config.MODEL.DROP_RATE,  # 0.0
        drop_path_rate=config.MODEL.DROP_PATH_RATE,  # can set to 0.1
        norm_layer=partial(nn.LayerNorm, eps=1e-6),
        init_values=config.MODEL.VIT.INIT_VALUES,
        use_abs_pos_emb=config.MODEL.VIT.USE_APE,  # False
        use_rel_pos_bias=config.MODEL.VIT.USE_RPB,   # True
        use_shared_rel_pos_bias=config.MODEL.VIT.USE_SHARED_RPB,  # False
        with_cls_token=config.MODEL.VIT.WITH_CLS_TOKEN,  # True
        use_checkpoint=config.TRAIN.USE_CHECKPOINT,)

    return model


def load_state_dict(path):
    checkpoint = torch.load(path, map_location='cpu')

    # rename moco pre-trained keys
    linear_keyword = 'head'
    state_dict = checkpoint
    if 'model' in checkpoint.keys():
        state_dict = checkpoint['model']
        
    for k in list(state_dict.keys()):
        # retain only base_encoder up to before the embedding layer
        if k.startswith('encoder'):
            # remove prefix
            new_key = k.replace('encoder.', '')
            state_dict[new_key] = state_dict.pop(k)
        elif k.startswith('feature_model') or k.startswith('decoder'):
            # remove these entries entirely
            state_dict.pop(k)

    return state_dict
